package net.sf.twip;

import java.lang.annotation.Annotation;
import java.util.*;
import java.util.concurrent.Callable;

import net.sf.twip.internal.*;
import net.sf.twip.util.ServiceLoadingException;

import org.junit.*;
import org.junit.runner.Description;
import org.junit.runners.BlockJUnit4ClassRunner;
import org.junit.runners.model.*;

public class ClassTwip extends BlockJUnit4ClassRunner {

	private List<FrameworkMethod> testMethods;

	private final List<TwipExtension> extensions = new ArrayList<TwipExtension>();

	private final Callable<?> testConstructor;

	private final Object[] parameterValues;

	public ClassTwip(Class<?> testClass, Callable<?> testConstructor, Object[] parameterValues)
			throws InitializationError {
		super(testClass);
		this.testConstructor = testConstructor;
		this.parameterValues = parameterValues;
		initExtensions(testClass);
	}

	@Override
	protected void collectInitializationErrors(List<Throwable> errors) {
		validatePublicVoidNoArgMethods(BeforeClass.class, true, errors);
		validatePublicVoidNoArgMethods(AfterClass.class, true, errors);
		// don't validate constructor!
		validateInstanceMethods(errors);
	}

	@Override
	protected Object createTest() throws Exception {
		return testConstructor.call();
	}

	@Override
	protected List<FrameworkMethod> computeTestMethods() {
		if (testMethods == null) {
			testMethods = new ArrayList<FrameworkMethod>();
			for (final FrameworkMethod method : getTestClass().getAnnotatedMethods(Test.class)) {
				if (method.getMethod().getParameterTypes().length == 0) {
					testMethods.add(method);
					continue;
				}
				List<? extends FrameworkMethod> parameterizedTestMethods;
				try {
					parameterizedTestMethods = FrameworkMethodWithArgs.of(method.getMethod());
				} catch (final TwipConfigurationError e) {
					parameterizedTestMethods = createFailedMethod(method, e);
				}
				testMethods.addAll(parameterizedTestMethods);
			}
		}
		return testMethods;
	}

	private List<FailedFrameworkMethod> createFailedMethod(final FrameworkMethod method,
			final Throwable e) {
		return Collections.singletonList(new FailedFrameworkMethod(method.getMethod(), e));
	}

	private void initExtensions(Class<?> testClass) throws InitializationError {
		final List<Throwable> errors = new ArrayList<Throwable>();

		final Set<Class<? extends TwipExtension>> loaded = instantiateServiceLoaderExtensions(errors);

		instantiateAnnotatedExtensions(testClass, loaded, errors);

		if (!errors.isEmpty()) {
			throw new InitializationError(errors);
		}
	}

	private Set<Class<? extends TwipExtension>> instantiateServiceLoaderExtensions(
			List<Throwable> errors) {
		final Set<Class<? extends TwipExtension>> serviceLoadedTypes = new HashSet<Class<? extends TwipExtension>>();
		final Iterator<TwipExtension> loaders = ServiceLoader.load(TwipExtension.class).iterator();
		while (loaders.hasNext()) {
			try {
				final TwipExtension extension = loaders.next();
				extensions.add(extension);
				serviceLoadedTypes.add(extension.getClass());
			} catch (final ServiceLoadingException e) {
				errors.add(e);
			}
		}
		return serviceLoadedTypes;
	}

	private void instantiateAnnotatedExtensions(Class<?> testClass,
			Set<Class<? extends TwipExtension>> loaded, final List<Throwable> errors) {
		final TwipExtensions annotation = testClass.getAnnotation(TwipExtensions.class);
		if (annotation != null) {
			for (final Class<? extends TwipExtension> type : annotation.value()) {
				if (loaded.contains(type))
					continue;
				try {
					extensions.add(type.newInstance());
				} catch (final InstantiationException e) {
					final String message = "can't instantiate " + type;
					errors.add(new TwipConfigurationErrorCantInstantiateExtension(message, e));
				} catch (final IllegalAccessException e) {
					errors.add(new TwipConfigurationErrorCantAccessExtension(
							"can't access " + type, e));
				}
			}
		}
	}

	@Override
	protected Statement methodInvoker(FrameworkMethod method, final Object testInstance) {
		Statement statement = super.methodInvoker(method, testInstance);
		for (final TwipExtension extension : extensions) {
			statement = extension.wrapInner(method, testInstance, statement);
		}
		return statement;
	}

	@Override
	protected Description describeChild(FrameworkMethod method) {
		String text = testName(method) + "(" + Arrays.toString(parameterValues) + ")";
		return Description.createSuiteDescription(text, method.getAnnotations());
	}

	@Override
	protected String testName(FrameworkMethod method) {
		return (method instanceof FrameworkMethodWithArgs) ? method.toString() : super
				.testName(method);
	}

	@Override
	protected String getName() {
		return (parameterValues == null) ? super.getName() : Arrays.toString(parameterValues);
	}

	@Override
	protected void validatePublicVoidNoArgMethods(Class<? extends Annotation> annotation,
			boolean isStatic, List<Throwable> errors) {
		final List<FrameworkMethod> methods = getTestClass().getAnnotatedMethods(annotation);
		for (final FrameworkMethod eachTestMethod : methods) {
			eachTestMethod.validatePublicVoid(isStatic, errors);
		}
	}

	@Override
	protected Statement withAfters(FrameworkMethod method, final Object testInstance,
			Statement statement) {
		Statement result = super.withAfters(method, testInstance, statement);
		for (final TwipExtension extension : extensions) {
			result = extension.wrapOuter(method, testInstance, result);
		}
		return result;
	}

	public Object[] getParameterValues() {
		return parameterValues;
	}
}
